{# Import the theme's layout. #}
{% extends "!page.html" %}

{%- block body %}

{%- if legacy_version -%}
  <a class="old-documentation" href="{{ pathto(master_doc) }}/../../{{ pagename }}{{ file_suffix }}">
    This documentation page refers to a previous release of DIALS ({{ legacy_version }}).<br/>
    Click here to go to the corresponding page for the latest version of DIALS
  </a>
{%- endif -%}

{{ super() }}

{%- endblock %}
