{# Import the theme's layout. #}
{% extends "!layout.html" %}

{% block header %}
  <div class="logoheader container">
  <a href="{{ pathto(master_doc) }}">
  <img class="logoheader" alt="DIALS" src="{{ pathto('_static/dials_header.png', 1) }}" />
  </a>
  </div>
  {{ super() }}
{% endblock %}


{% block relbar1 %}{% endblock %}

{% block footer %}
  <div class="footer container">
  <a href="https://www.diamond.ac.uk/"><img class="logofooter" alt="Diamond" src="{{ pathto('_static/diamond_logo.png', 1) }}" /></a>
  <a href="https://www.ccp4.ac.uk/"><img class="logofooter" alt="CCP4" src="{{ pathto('_static/CCP4-logo-plain.png', 1) }}" /></a>
  <a href="https://www.stfc.ac.uk/"><img class="logofooter" alt="STFC" src="https://dials.github.io/images/logos/ukri-stfc.png" /></a>
  <a href="https://www.lbl.gov/"><img class="logofooter" alt="LBL" src="https://dials.github.io/images/logos/lbl.png" /></a>
  <a href="http://smb.slac.stanford.edu/"><img class="logofooter" alt="SSRL/SMB" src="https://dials.github.io/images/logos/smbssrl.jpg" /></a>
  </div>

  <script type="text/javascript">
     $(document).ready(function() {
         $(".toggle > *").hide();
         $(".toggle .header").show();
         $(".toggle .header").click(function() {
             $(this).parent().children().not(".header").toggle(400);
             $(this).parent().children(".header").toggleClass("open");
         })
     });
  </script>
  {{ super() }}
{% endblock %}

{# Custom CSS overrides #}
{% set bootswatch_css_custom = ['_static/dials-styles.css'] %}

{% set css_files = css_files + ['_static/dials-styles.css'] %}
