{% extends "report_base.html" %}
{% block content %}
<div class="page-header">
    <h1>{{ page_title }}</h1>
</div>

<div>
  <div class="panel-group">
      {{ macros.panel(panel_title, panel_id, graphs, show=True) }}
  </div>
</div>

{% endblock %}
