/*
 * mask_code.h
 *
 *  Copyright (C) 2013 Diamond Light Source
 *
 *  Author: James Parkhurst
 *
 *  This code is distributed under the BSD license, a copy of which is
 *  included in the root directory of this package.
 */
#ifndef DIALS_ALGORITHMS_SHOEBOX_MASK_CODE_H
#define DIALS_ALGORITHMS_SHOEBOX_MASK_CODE_H

#include <dials/model/data/shoebox.h>

namespace dials { namespace algorithms { namespace shoebox {

  using dials::model::Background;
  using dials::model::BackgroundUsed;
  using dials::model::Foreground;
  using dials::model::MaskCode;
  using dials::model::Overlapped;
  using dials::model::Strong;
  using dials::model::Valid;

}}};  // namespace dials::algorithms::shoebox

#endif /* DIALS_ALGORITHMS_SHOEBOX_MASK_CODE_H */
