from __future__ import annotations

from dials_algorithms_polygon_clip_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "line_with_rect",
    "quad_with_convex_quad",
    "quad_with_triangle",
    "simple_with_convex",
    "simple_with_rect",
    "triangle_with_convex_quad",
    "triangle_with_triangle",
)
