#ifndef DIALS_ARRAY_FAMILY_SCITBX_SHARED_AND_VERSA_H
#define DIALS_ARRAY_FAMILY_SCITBX_SHARED_AND_VERSA_H

#include <scitbx/array_family/versa.h>
#include <scitbx/array_family/shared.h>
#include <scitbx/array_family/ref.h>
#include <scitbx/array_family/accessors/c_grid.h>
#include <dials/array_family/import_scitbx_af.h>

#endif  // DIALS_ARRAY_FAMILY_SCITBX_SHARED_AND_VERSA_H
