#ifndef DIALS_ARRAY_FAMILY_IMPORT_SCITBX_AF_H
#define DIALS_ARRAY_FAMILY_IMPORT_SCITBX_AF_H

namespace dials { namespace af {
  using namespace scitbx::af;
}}  // namespace dials::af

#endif  // DIALS_ARRAY_FAMILY_IMPORT_SCITBX_AF_H
